# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - ImprovedTableParser library

    @copyright: 2012 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from MoinMoin import wikiutil
from shlex import shlex
from StringIO import StringIO
import re

# Regular expressions.

syntax = {
    # For section markers.
    "markers"   : (r"^\s*(?P<n>\\+)(?P<b>{|})(?P=n)(?P=b)(?P=n)(?P=b)", re.MULTILINE),
    "marker"    : (r"(\\+)",                                            0),

    # At start of line:
    "sections"  : (r"(^\s*{{{.*?^\s*}}})",  re.MULTILINE | re.DOTALL),  # {{{ ... }}}
    "rows"      : (r"^==",                  re.MULTILINE),              # ==

    # Within text:
    "columns"   : (r"\|\|[ \t]*",           0),                         # || ws-excl-nl

    # At start of column text:
    "column"    : (r"^\s*<(.*?)>\s*(.*)",   re.DOTALL),                 # ws < attributes > ws
    }

patterns = {}
for name, (value, flags) in syntax.items():
    patterns[name] = re.compile(value, re.UNICODE | flags)

# Other regular expressions.

leading_number_regexp = re.compile(r"\d*")

# Functions.

def parse(s):

    "Parse 's', returning a table definition."

    s = replaceMarkers(s)

    table_attrs = {}
    rows = []

    # The following will be redefined upon the construction of the first row.

    row_attrs = {}
    columns = []

    # Process exposed text and sections.

    exposed = True

    # Initially, start a new row.

    row_continued = False

    for region in patterns["sections"].split(s):

        # Only look for table features in exposed text.

        if exposed:

            # Extract each row from the definition.

            for row_text in patterns["rows"].split(region):

                # Only create a new row when a boundary has been found.

                if not row_continued:
                    if columns:
                        extractAttributes(columns[0][0], row_attrs, table_attrs)

                    row_attrs = {}
                    columns = []
                    rows.append((row_attrs, columns))
                    column_continued = False

                # Extract each column from the row.

                for text in patterns["columns"].split(row_text):

                    # Only create a new column when a boundary has been found.

                    if not column_continued:

                        # Extract the attribute and text sections.

                        match = patterns["column"].search(text)
                        if match:
                            attribute_text, text = match.groups()
                            columns.append([parseAttributes(attribute_text, True), text])
                        else:
                            columns.append([{}, text])

                    else:
                        columns[-1][1] += text

                    # Permit columns immediately following this one.

                    column_continued = False

                # Permit a continuation of the current column.

                column_continued = True

                # Permit rows immediately following this one.

                row_continued = False

            # Permit a continuation if the current row.

            row_continued = True

        # Write any section into the current column.

        else:
            columns[-1][1] += region

        exposed = not exposed

    if columns:
        extractAttributes(columns[0][0], row_attrs, table_attrs)

    return table_attrs, rows

def extractAttributes(attrs, row_attrs, table_attrs):

    """
    Extract row- and table-level attributes from 'attrs', storing them in
    'row_attrs' and 'table_attrs' respectively.
    """

    for name, value in attrs.items():
        if name.startswith("row") and name != "rowspan":
            row_attrs[name] = value
            del attrs[name]
        elif name.startswith("table"):
            table_attrs[name] = value
            del attrs[name]

def replaceMarkers(s):

    "Convert the section notation in 's'."

    l = []
    last = 0

    # Get each marker and convert it.

    for match in patterns["markers"].finditer(s):
        start, stop = match.span()
        l.append(s[last:start])

        # Convert the marker.

        marker = []
        brace = True
        for text in patterns["marker"].split(match.group()):
            if brace:
                marker.append(text)
            else:
                marker.append(text[:-1])
            brace = not brace

        l.append("".join(marker))
        last = stop
    else:
        l.append(s[last:])

    return "".join(l)

def parseAttributes(s, escape=True):

    """
    Parse the table attributes string 's', returning a mapping of names to
    values. If 'escape' is set to a true value, the attributes will be suitable
    for use with the formatter API. If 'escape' is set to a false value, the
    attributes will have any quoting removed.
    """

    attrs = {}
    f = StringIO(s)
    name = None
    need_value = False

    for token in shlex(f):

        # Capture the name if needed.

        if name is None:
            name = escape and wikiutil.escape(token) or strip_token(token)

        # Detect either an equals sign or another name.

        elif not need_value:
            if token == "=":
                need_value = True
            else:
                attrs[name.lower()] = escape and "true" or True
                name = wikiutil.escape(token)

        # Otherwise, capture a value.

        else:
            # Quoting of attributes done similarly to wikiutil.parseAttributes.

            if token:
                if escape:
                    if token[0] in ("'", '"'):
                        token = wikiutil.escape(token)
                    else:
                        token = '"%s"' % wikiutil.escape(token, 1)
                else:
                    token = strip_token(token)
    
            attrs[name.lower()] = token
            name = None
            need_value = False

    return attrs

def strip_token(token):

    "Return the given 'token' stripped of quoting."

    if token[0] in ("'", '"') and token[-1] == token[0]:
        return token[1:-1]
    else:
        return token

# Formatting of embedded content.
# NOTE: Borrowed from EventAggregator.

def getParserClass(request, format):

    """
    Return a parser class using the 'request' for the given 'format', returning
    a plain text parser if no parser can be found for the specified 'format'.
    """

    try:
        return wikiutil.searchAndImportPlugin(request.cfg, "parser", format or "plain")
    except wikiutil.PluginMissingError:
        return wikiutil.searchAndImportPlugin(request.cfg, "parser", "plain")

def formatText(text, request, fmt):

    "Format the given 'text' using the specified 'request' and formatter 'fmt'."

    parser_cls = getParserClass(request, request.page.pi["format"])
    parser = parser_cls(text, request, line_anchors=False)
    return request.redirectedOutput(parser.format, fmt, inhibit_p=True)

# Sorting utilities.

def get_sort_columns(s, start=0):

    """
    Split the comma-separated string 's', extracting the column specifications
    of the form <column>["n"] where the prefix "n" indicates an optional
    numeric conversion for that column. Column indexes start from the specified
    'start' value (defaulting to 0).
    """

    sort_columns = []
    for column_spec in s.split(","):
        column_spec = column_spec.strip()

        ascending = True
        if column_spec.endswith("d"):
            column_spec = column_spec[:-1]
            ascending = False

        # Extract the conversion indicator and column index.

        if column_spec.endswith("n"):
            column = column_spec[:-1]
            fn = to_number
        else:
            column = column_spec
            fn = str

        # Ignore badly-specified columns.

        try:
            sort_columns.append((max(0, int(column) - start), fn, ascending))
        except ValueError:
            pass

    return sort_columns

def to_number(s):

    "Convert 's' to a number, discarding any non-numeric trailing data."

    match = leading_number_regexp.match(s)
    if match:
        return int(match.group())
    else:
        raise ValueError, s

class Sorter:

    "A sorting helper class."

    def __init__(self, sort_columns):
        self.sort_columns = sort_columns

    def __call__(self, row1, row2):
        row_attrs1, columns1 = row1
        row_attrs2, columns2 = row2

        # Apply the conversions to each column, comparing the results.

        for column, fn, ascending in self.sort_columns:
            column_attrs1, text1 = columns1[column]
            column_attrs2, text2 = columns2[column]

            # Ignore a column when a conversion is not possible.

            try:
                text1 = fn(text1)
                text2 = fn(text2)
                result = cmp(text1, text2)

                # Where the columns differ, return a result observing the sense
                # (ascending or descending) of the comparison for the column.

                if result != 0:
                    return ascending and result or -result

            except ValueError:
                pass

        return 0

# Common formatting functions.

def formatTable(text, request, fmt, attrs=None):

    """
    Format the given 'text' using the specified 'request' and formatter 'fmt'.
    The optional 'attrs' can be used to control the presentation of the table.
    """

    table_attrs, table = parse(text)

    # Sort the rows according to the values in each of the specified columns.

    if attrs.has_key("sortcolumns"):
        data_start = int(attrs.get("headers", "1"))
        headers = table[:data_start]
        data = table[data_start:]

        # Get the sort columns using Unix sort-like notation.

        sorter = Sorter(get_sort_columns(attrs["sortcolumns"]))
        data.sort(cmp=sorter)

        table = headers + data

    # Write the table.

    request.write(fmt.table(1, table_attrs))

    for row_attrs, columns in table:
        request.write(fmt.table_row(1, row_attrs))

        for column_attrs, column_text in columns:
            request.write(fmt.table_cell(1, column_attrs))
            request.write(formatText(column_text, request, fmt))
            request.write(fmt.table_cell(0))

        request.write(fmt.table_row(0))

    request.write(fmt.table(0))

# vim: tabstop=4 expandtab shiftwidth=4
